import logging
from datetime import timedelta
from enum import Enum, auto
from typing import List, Optional

from attr import define

from automsr.config import Profile

logger = logging.getLogger(__name__)


class OutcomeType(Enum):
    """
    Possible outcome for an execution.
    """

    FAILURE = "failure"
    SUCCESS = "success"
    SKIPPED = "skipped"


class StepType(Enum):
    """
    Type of execution step found during the normal execution.
    """

    START_SESSION = auto()
    END_SESSION = auto()

    GET_DASHBOARD = auto()
    GET_POINTS = auto()

    PUNCHCARDS = auto()
    PROMOTIONS = auto()
    PC_SEARCHES = auto()
    MOBILE_SEARCHES = auto()

    @classmethod
    def get_ordered_steps(cls) -> List["StepType"]:
        """
        Returns the step ordered as expected by the `run` flow.
        """

        return [
            # start the session
            cls.START_SESSION,
            # retrieve the dashboard
            cls.GET_DASHBOARD,
            # complete the punchcards, if any
            cls.PUNCHCARDS,
            # complete the promotions, including daily
            cls.PROMOTIONS,
            # complete searches
            cls.PC_SEARCHES,
            cls.MOBILE_SEARCHES,
            # get the total points
            cls.GET_POINTS,
            # end the session
            cls.END_SESSION,
        ]


@define
class Step:
    """
    Status of a single step of an entire execution.
    """

    type: StepType
    outcome: OutcomeType
    explanation: Optional[str] = None
    duration: Optional[timedelta] = None


@define
class Status:
    """
    Status of the whole execution for a profile.
    """

    profile: Profile
    steps: List[Step]
    points: Optional[int]

    def get_outcome(self) -> OutcomeType:
        """
        Find the overall outcome based on the outcome of the children steps.

        >>> from unittest.mock import ANY
        >>> profile = Profile(email="foo@bar.com", profile="profile 1")
        >>> success = Step(type=ANY, outcome=OutcomeType.SUCCESS)
        >>> failure = Step(type=ANY, outcome=OutcomeType.FAILURE)
        >>> skipped = Step(type=ANY, outcome=OutcomeType.SKIPPED)

        >>> Status(profile=profile, steps=[], points=None).get_outcome().name
        'SKIPPED'
        >>> Status(profile=profile, steps=[success], points=None).get_outcome().name
        'SUCCESS'
        >>> Status(profile=profile, steps=[success, skipped], points=None).get_outcome().name
        'SUCCESS'
        >>> Status(profile=profile, steps=[skipped, skipped], points=None).get_outcome().name
        'SKIPPED'
        >>> Status(profile=profile, steps=[skipped, skipped, failure], points=None).get_outcome().name
        'FAILURE'
        """

        outcomes = [step.outcome for step in self.steps]

        # If no step is present, the status defaults to SKIPPED.
        if not outcomes:
            return OutcomeType.SKIPPED

        # The outcome is SKIPPED if all the outcomes are SKIPPED.
        if all(outcome is OutcomeType.SKIPPED for outcome in outcomes):
            return OutcomeType.SKIPPED

        # The outcome is SUCCESS if no outcome is FAILURE,
        # so they are either SUCCESS or SKIPPED.
        if not any(outcome is OutcomeType.FAILURE for outcome in outcomes):
            return OutcomeType.SUCCESS

        # If we are here, at least one outcome is FAILURE.
        return OutcomeType.FAILURE



import logging
import time
from datetime import timedelta
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union, cast

import selenium.common.exceptions
from attr import define, field
from selenium.common import NoSuchElementException
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from tqdm import tqdm

from automsr.browser.browser import Browser
from automsr.config import Config, Profile
from automsr.datatypes.dashboard import (
    Dashboard,
    Promotion,
    PromotionType,
    QuizType,
)
from automsr.datatypes.execution import (
    OutcomeType,
    Status,
    Step,
    StepType,
)
from automsr.mail import EmailExecutor
from automsr.search import RandomSearchGenerator

logger = logging.getLogger(__name__)


class ExecutorException(Exception):
    """
    Base class for Executor exceptions.
    """


class QuizException(ExecutorException):
    """
    Base class for Quiz Promotions exceptions.
    """


class CannotFindStartButtonException(QuizException):
    """
    Exception raised when the Start button for quizzes is expected but not found.
    """


class CannotDetermineQuizTypeException(QuizException):
    """
    Exception raised when the Executor is unable to determine
    the quiz-type of a promotion.
    """


def get_available_points(dashboard: Dashboard) -> int:
    """
    Return the available points that a user can have.
    """

    return dashboard.points()


@define
class SingleTargetExecutor:
    """
    Executor class driving the completion of all the tasks associated with Rewards.

    This executor will target a single profile.
    """

    config: Config
    profile: Profile
    browser: Browser = field(init=False)

    def _get_execution_function(
        self, step_type: StepType, dashboard: Optional[Dashboard] = None
    ) -> Callable[[], Optional[Union[Dashboard, int]]]:
        """
        Return the function associated to the input execution step.
        """

        func: Callable

        if step_type is StepType.START_SESSION:
            return self.start_session

        elif step_type is StepType.GET_DASHBOARD:
            func = self.get_dashboard

        elif step_type in (
            StepType.PROMOTIONS,
            StepType.PUNCHCARDS,
            StepType.PC_SEARCHES,
            StepType.MOBILE_SEARCHES,
            StepType.GET_POINTS,
        ):
            assert dashboard is not None

            if step_type is StepType.PROMOTIONS:
                func = partial(self.execute_promotions, dashboard=dashboard)
            elif step_type is StepType.PUNCHCARDS:
                func = partial(self.execute_punchcards, dashboard=dashboard)
            elif step_type is StepType.PC_SEARCHES:
                func = partial(self.execute_pc_searches, dashboard=dashboard)
            elif step_type is StepType.MOBILE_SEARCHES:
                func = partial(self.execute_mobile_searches, dashboard=dashboard)
            elif step_type is StepType.GET_POINTS:
                func = partial(get_available_points, dashboard=dashboard)
            else:
                raise ValueError(step_type)

        elif step_type is StepType.END_SESSION:
            func = self.end_session

        else:
            raise ValueError(step_type)

        return func

    def execute(self) -> Status:
        """
        Execute the following steps:
        - Open a new browser session with Selenium and a Chrome driver.
        - Retrieval of dashboard json from Rewards page
        - Execution of all completable punchcards
        - Execution of all completable promotions
        - Execution of searches:
            - PC searches (desktop user agent)
            - Mobile searches (mobile user agent)

        Returns the Status of the execution for the current profile.
        """

        # Construct the list of statuses for each step.
        steps: List[Step] = []

        # Get the list of steps in the correct order of execution
        step_types: List[StepType] = StepType.get_ordered_steps()

        # If the profile is marked as skipped, return this result.
        if self.profile.skip:
            logger.info("Profile marked as skipped.")
            steps = [
                Step(type=step_type, outcome=OutcomeType.SKIPPED)
                for step_type in step_types
            ]
            status = Status(
                profile=self.profile,
                steps=steps,
                points=None,
            )
            return status

        # Declare the variables needed in the following loop.
        dashboard: Optional[Dashboard] = None
        points: Optional[int] = None

        for step_type in step_types:
            logger.debug("Step type under execution: %s", step_type)

            # Get the function to use.
            function = self._get_execution_function(
                step_type=step_type, dashboard=dashboard
            )

            # Declare the outcome of the step.
            outcome: Optional[OutcomeType] = None

            # Declare the explanation of the step, if non-success.
            explanation: Optional[str] = None

            # Snapshot a performance counter and returns the value in seconds.
            start_counter: float = time.perf_counter()

            try:
                # Execute the function.
                retval = function()

                # If the step is get-dashboard, we will expect a Dashboard return value.
                if step_type is StepType.GET_DASHBOARD:
                    dashboard = cast(Dashboard, retval)
                    assert isinstance(dashboard, Dashboard)

                # If the step is get-points, we will expect an integer return value.
                elif step_type is StepType.GET_POINTS:
                    points = cast(int, retval)
                    assert isinstance(points, int)

                # Otherwise, we expect nothing as return value.
                else:
                    assert retval is None

            # Catch exceptions block.
            # Selenium errors will be treated as normal errors, but won't fail the entire execution;
            #  only the corresponding step will be marked as failed.
            except selenium.common.exceptions.WebDriverException as e:
                outcome = OutcomeType.FAILURE
                explanation = str(e)
                logger.error("Selenium exception caught: %s", e)

            # This is needed for Punchcards.
            except NotImplementedError:
                outcome = OutcomeType.SKIPPED
                explanation = "Step not implemented yet."
                logger.warning(explanation)

            # If no exception is raised, the step is successful.
            # Otherwise, any uncaught exception will be propagated to the caller.
            else:
                outcome = OutcomeType.SUCCESS
                logger.debug("Step completed correctly.")

            # At the end of the try-except block, we will have an outcome, a duration,
            # and possibly an explanation.
            finally:
                # Capture time elapsed.
                end_counter = time.perf_counter()
                seconds = end_counter - start_counter
                duration = timedelta(seconds=seconds)

                # Create step based on step status.
                assert outcome is not None
                step = Step(
                    type=step_type,
                    outcome=outcome,
                    duration=duration,
                    explanation=explanation,
                )

                # Append the created step status to the list of statuses.
                steps.append(step)

        # Create the overall status for the current profile, and return it to the caller.
        status = Status(profile=self.profile, steps=steps, points=points)
        return status

    def start_session(self) -> None:
        """
        Create a new session with Selenium and Chromedriver,
        then returns the session object to the caller.
        """

        self.browser = Browser.from_config(config=self.config, profile=self.profile)
        self.browser.test_driver()
        self.browser.go_to(self.config.automsr.rewards_homepage)

        # Simulate the user landing on the page and waiting a little bit
        time.sleep(2)

    def end_session(self) -> None:
        """
        Terminate the session.
        """

        self.browser.driver.quit()

    def get_dashboard(self) -> Dashboard:
        """
        Retrieve a Dashboard object from the current page.

        This method is expected to be run inside the Rewards homepage.
        """

        raw_data: Dict[str, Any] = self.browser.execute_script("return dashboard;")
        dashboard = Dashboard(**raw_data)
        return dashboard

    def execute_pc_searches(self, dashboard: Dashboard) -> None:
        """
        Execute PC searches, if needed.
        """

        if (amount := dashboard.amount_of_pc_searches()) == 0:
            logger.info("No PC search is needed")
            return
        else:
            logger.info("Starting PC searches")

        assert amount > 0
        return self._execute_searches(
            amount=amount, user_agent=self.browser.user_agents.desktop
        )

    def execute_mobile_searches(self, dashboard: Dashboard) -> None:
        """
        Execute Mobile searches, if needed.
        """

        if (amount := dashboard.amount_of_mobile_searches()) == 0:
            logger.info("No Mobile search is needed")
            return
        else:
            logger.info("Starting Mobile searches")

        assert amount > 0
        return self._execute_searches(
            amount=amount, user_agent=self.browser.user_agents.mobile
        )

    def _execute_searches(
        self, amount: int = 1, user_agent: Optional[str] = None
    ) -> None:
        """
        Helper method to execute `amount` searches.

        Can specify a custom `user_agent` to use.
        """

        assert amount >= 1, "Invalid value!"

        safe_amount = amount + 5
        logger.debug("Original amount of searches: %s", amount)
        logger.debug("Safe amount of searches: %s", safe_amount)

        old_user_agent: Optional[str] = None
        if user_agent is not None:
            logger.debug("Changing user-agent to: %s", user_agent)
            old_user_agent = self.browser.get_user_agent()
            self.browser.change_user_agent(user_agent=user_agent)

        search_generator = RandomSearchGenerator()
        sleep_time = search_generator.sleep_time()
        query = search_generator.query_gen()

        self.browser.go_to_bing()

        for i in tqdm(range(safe_amount)):
            logger.debug("Executing search: %s/%s", i + 1, safe_amount)

            # we retrieve the element in the for-loop since the page is reloaded, thus the element can be invalidated
            element: WebElement = self.browser.driver.find_element(
                by=By.ID, value="sb_form_q"
            )

            # send the next item of the generator to the input field
            element.send_keys(next(query))

            # sleep to prevent issues with Selenium interacting with the page
            time.sleep(0.5)

            # send ENTER to perform a search
            element.send_keys(Keys.ENTER)

            # sleep a certain amount of time
            time.sleep(sleep_time)

        if user_agent is not None:
            logger.debug("Restoring original user-agent: %s", old_user_agent)
            assert old_user_agent is not None
            self.browser.change_user_agent(user_agent=old_user_agent)

    def execute_promotions(self, dashboard: Dashboard) -> None:
        """
        Execute all completable promotions.
        """

        # handle promotional items differently, since there is another way
        # to retrieve the corresponding WebElement
        promotional_item = dashboard.get_promotional_item()
        if promotional_item is not None:
            try:
                self.browser.open_promotion(
                    promotion=promotional_item, element_id="promo-item"
                )
            except Exception as e:
                logger.error("Exception caught: %s", e)
                logger.warning("Promotion '%s' will be skipped", promotional_item)
            finally:
                # simulate navigation
                logger.debug("Re-opening Rewards homepage to simulate navigation.")
                self.browser.go_to_rewards()
                logger.debug("Sleeping to simulate a real user behavior.")
                time.sleep(2)

        # handle all other promotions
        promotions: List[Promotion] = dashboard.get_completable_promotions()

        for promotion in promotions:
            logger.info("Executing promotion: %s", promotion.title)
            try:
                self._execute_promotion(promotion=promotion)
            except Exception as e:
                logger.error("Exception caught: %s", e)
                logger.warning("Promotion '%s' will be skipped", promotion)
            finally:
                # simulate navigation
                logger.debug("Re-opening Rewards homepage to simulate navigation.")
                self.browser.go_to_rewards()
                logger.debug("Sleeping to simulate a real user behavior.")
                time.sleep(2)

    def _execute_promotion(self, promotion: Promotion) -> None:
        """
        Execute a specific promotion.
        """

        driver = self.browser.driver
        self.browser.open_promotion(promotion=promotion)

        # Sleep to simulate user behavior and to let JS load the page
        time.sleep(3)

        # Determine which promotion is currently on
        if promotion.promotionType == PromotionType.QUIZ:
            # Try to determine which quiz we are dealing with
            try:
                _ = driver.find_element(by=By.ID, value="btoption0")
            except NoSuchElementException:
                pass
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.CHOICE_BETWEEN_TWO
                )

            # If the quiz is not a choice-between-two, try with the three-questions-N-answers
            try:
                start_button = driver.find_element(by=By.ID, value="rqStartQuiz")
                # if we find the button, we click it and then wait a little bit for JS
                # to load the answers in the DOM
                start_button.click()
                time.sleep(2)
            except NoSuchElementException as e:
                # if we are here, the quiz can be either already started,
                # or something's broken; in the latter case, we raise an error
                try:
                    driver.find_element(by=By.ID, value="currentQuestionContainer")
                except NoSuchElementException:
                    logger.error(
                        "Cannot find neither the start button, nor the quiz container!"
                    )
                    raise CannotFindStartButtonException() from e
                else:
                    logger.info("Quiz already started.")

            # Check if the quiz is 8-answers
            try:
                _ = driver.find_element(by=By.ID, value="rqAnswerOption7")
            except NoSuchElementException:
                pass
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.THREE_QUESTIONS_EIGHT_ANSWERS
                )

            # Check if the quiz is 4-answers
            try:
                _ = driver.find_element(by=By.ID, value="rqAnswerOption3")
            except NoSuchElementException as e:
                raise CannotDetermineQuizTypeException() from e
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.THREE_QUESTIONS_FOUR_ANSWERS
                )

        elif promotion.promotionType in (
            PromotionType.URL_REWARD,
            PromotionType.WELCOME_TOUR,
        ):
            logger.info("Promotion executed by just opening the destination url.")
        else:
            raise ValueError(
                f"Cannot execute promotion with type: {promotion.promotionType}"
            )

    def _execute_quiz_promotion(self, quiz_type: QuizType) -> None:
        """
        Execute a specific promotion that is a quiz with a known type.

        The method assumes that, if a quiz type requires for a Start button to be clicked,
        this is done by the caller.
        """

        driver = self.browser.driver

        logger.info("Resolving promotion as quiz with type: %s", quiz_type)

        if quiz_type is QuizType.CHOICE_BETWEEN_TWO:
            button_id = "btoption0"

            # safety breaks
            loop_break = 5
            loop_counter = 1

            while loop_counter < loop_break:
                button = driver.find_element(by=By.ID, value=button_id)
                _class = button.get_attribute("class")
                is_quiz_completed = _class is not None and "selectedOption" in _class

                if is_quiz_completed:
                    logger.info("Quiz finished")
                    return
                else:
                    button.click()

                    # Wait some seconds to let the page reload
                    time.sleep(1)

                loop_counter += 1

            # if we are here, we didn't finish the quiz
            logger.warning("Quiz not finished!")
            return

        elif quiz_type in (
            QuizType.THREE_QUESTIONS_FOUR_ANSWERS,
            QuizType.THREE_QUESTIONS_EIGHT_ANSWERS,
        ):
            # safety breaks
            loop_break = 100
            loop_counter = 1

            # Create the `max_answer_index` based on the provided quiz type.
            if quiz_type is QuizType.THREE_QUESTIONS_FOUR_ANSWERS:
                max_answer_index = 4
            elif quiz_type is QuizType.THREE_QUESTIONS_EIGHT_ANSWERS:
                max_answer_index = 8
            else:
                raise ValueError(f"Invalid quiz type provided: {quiz_type}")

            current_answer_index = 0

            while loop_counter < loop_break:
                # check if quiz is finished
                try:
                    driver.find_element(by=By.ID, value="quizCompleteContainer")
                except NoSuchElementException:
                    logger.debug("Quiz still in progress.")
                else:
                    logger.info("Quiz finished.")
                    return

                # if not, continue with clicking answers
                answer_id = f"rqAnswerOption{current_answer_index}"
                answer_element = driver.find_element(by=By.ID, value=answer_id)
                answer_element.click()

                # update the answer index
                current_answer_index = (current_answer_index + 1) % max_answer_index

                # update the loop counter
                loop_counter += 1

            # if we are here, we didn't finish the quiz
            logger.warning("Quiz not finished!")
            return
        else:
            raise ValueError(f"Quiz type not supported: {quiz_type}")

    def execute_punchcards(self, dashboard: Dashboard) -> None:
        """
        Execute all completable Punchcards, if any.
        """

        raise NotImplementedError


@define
class MultipleTargetsExecutor:
    """
    Executor class driving the completion of all the tasks associated with Rewards.

    This executor will target multiple profiles automatically.
    """

    config: Config

    def execute(self) -> None:
        """
        Spawn a SingleTargetExecutor for every profile specified in the `config` file.
        """

        statuses: List[Status] = []

        for profile in self.config.automsr.profiles:
            logger.info("Profile under execution: %s", profile)
            executor = SingleTargetExecutor(config=self.config, profile=profile)
            status = executor.execute()
            logger.info("Status of execution for profile %s: %s", profile, status)
            statuses.append(status)

        email_executor = EmailExecutor(config=self.config)
        if email_executor.are_messages_enabled():
            email_executor.send_message(statuses=statuses)



import datetime
import logging
import random
import smtplib
from email.message import EmailMessage
from typing import Dict, List, Optional

import markdown
from attr import define, field
from faker import Faker  # type: ignore

from automsr.config import Config
from automsr.datatypes.execution import OutcomeType, Status, Step, StepType

logger = logging.getLogger(__name__)


@define
class StatusMessage:
    """
    Wrapper for a status object to make it messages-aware.
    """

    status: Status

    @property
    def email(self) -> str:
        """
        Email address related to this object.
        """

        return self.status.profile.email

    def to_plain_text(self) -> str:
        """
        Return a plain-text representation of the status.
        """

        points_str = "N/A" if self.status.points is None else f"{self.status.points:,}"

        retval: List[str] = [
            f"Email: {self.email}",
            f"Points: {points_str}",
            f"Overall outcome: {self.status.get_outcome().name}",
        ]

        steps = self.status.steps
        for step in steps:
            line = f"Step {step.type.name} has outcome {step.outcome.name}."
            if step.duration:
                line += f" Duration: {step.duration}."
            if step.explanation:
                line += f" Explanation: {step.explanation}"
            retval.append(line)

        message = "\n".join(retval)
        message += "\n"  # add a final newline character
        return message

    def to_markdown(self) -> str:
        """
        Return a Markdown representation of the status.
        """

        status_emojis: Dict[OutcomeType, str] = {
            OutcomeType.SUCCESS: "✔️",
            OutcomeType.FAILURE: "❌",
            OutcomeType.SKIPPED: "⏭️",
        }

        overall_outcome = self.status.get_outcome()
        overall_outcome_emoji = status_emojis[overall_outcome]

        points_str = "N/A" if self.status.points is None else f"{self.status.points:,}"

        retval: List[str] = [
            f"### Profile: {self.email}",
            f"**Points: {points_str}**",
            "",
            f"**Overall outcome: {overall_outcome_emoji} {overall_outcome.name}**",
            # Table creation
            "",
            "| Outcome | Step | Duration | Explanation |",
            "| :---: | :--- | :--- | :--- |",
        ]

        for step in self.status.steps:
            outcome_emoji = status_emojis[step.outcome]
            explanation = step.explanation or ""
            duration = step.duration or "N/A"
            line = (
                f"| {outcome_emoji} | {step.type.name} | {duration} | {explanation} |"
            )
            retval.append(line)

        return "\n".join(retval)

    def to_html(self) -> str:
        """
        Return an HTML representation of the status.
        """

        return markdown.markdown(text=self.to_markdown(), extensions=["extra"])


@define
class ExecutionMessage:
    """
    Email message wrapper, customized for the tool logic.
    """

    sender: str
    recipient: str
    status_messages: List[StatusMessage]
    subject: str = field(
        factory=lambda: f"AutoMSR {datetime.date.today()} Report message"
    )

    def get_message(self) -> EmailMessage:
        """
        Returns a valid Email Message complaint with RFC 5322.

        >>> _message = ExecutionMessage(
        ...     sender="sender@example.com",
        ...     recipient="recipient@example.com",
        ...     status_messages=[],
        ...     subject="My Subject",
        ... ).get_message()
        >>> print(_message.as_string())  # doctest: +ELLIPSIS
        From: sender@example.com
        To: recipient@example.com
        Subject: My Subject
        MIME-Version: 1.0
        ...

        >>> from unittest.mock import patch
        >>> with patch("automsr.mail.datetime", wraps=datetime) as mock:
        ...     mock.date.today.return_value = "2023-01-01"
        ...     _message = ExecutionMessage(
        ...         sender="sender@example.com",
        ...         recipient="recipient@example.com",
        ...         status_messages=[],
        ...     ).get_message()
        ...     print(_message.as_string())  # doctest: +ELLIPSIS
        From: sender@example.com
        To: recipient@example.com
        Subject: AutoMSR 2023-01-01 Report message
        MIME-Version: 1.0
        ...
        """

        message = EmailMessage()

        message.add_header("From", self.sender)
        message.add_header("To", self.recipient)
        message.add_header("Subject", self.subject)

        complete_message_plain_text = "\n\n".join(
            [message.to_plain_text() for message in self.status_messages]
        )
        complete_message_html = "<br>".join(
            [message.to_html() for message in self.status_messages]
        )

        message.set_content(complete_message_plain_text)
        message.add_alternative(complete_message_html, subtype="html")

        return message


@define
class EmailConnection:
    sender: str
    password: str
    host: str
    port: int = field(converter=int)
    tls: bool = field(default=False, converter=bool)
    smtp: smtplib.SMTP = field(init=False)

    def open(self) -> None:
        """
        Open an SMTP connection and try to log in with mail server.
        """

        self.smtp = smtplib.SMTP(host=self.host, port=self.port)
        self.smtp.ehlo()
        if self.tls:
            self.smtp.starttls()
        self.smtp.login(self.sender, self.password)

    def close(self) -> None:
        """
        Closes an SMTP established connection.
        """

        try:
            self.smtp.quit()
        except smtplib.SMTPServerDisconnected:
            pass
        finally:
            logger.debug("SMTP connection closed")

    def test_connection(self) -> None:
        """
        Test if the provided connection is successful.
        Any exception is raised to the caller.
        """

        self.open()
        self.close()
        logger.info("Email connection was successful")

    def send_message(self, message: ExecutionMessage) -> None:
        """
        Send a Message using the provided credentials and SMTP.
        """
        smtp_message = message.get_message()
        self.smtp.send_message(msg=smtp_message)
        logger.debug("Message sent correctly with smtp!")

    def __enter__(self):
        return self.open()

    def __exit__(self, _exc_type, _exc_value, _traceback):
        # Check docs for more info on these parameters:
        # https://docs.python.org/3/reference/datamodel.html#object.__exit__
        self.close()


@define
class EmailConnectionFactory:
    """
    Connection factory; will try to create automagically a connection based on the sender address.
    """

    config: Config

    gmail_host = "smtp.gmail.com"
    gmail_port = 587
    gmail_tls = True

    outlook_host = "smtp-mail.outlook.com"
    outlook_port = 587
    outlook_tls = True

    def get_connection(self) -> Optional[EmailConnection]:
        """
        Get an email connection based on the config.

        >>> from pathlib import Path
        >>> current_path = Path(__file__)
        >>> tests_path = current_path.parent.parent / "tests"
        >>> config_path = tests_path / "configs/config.example.yaml"
        >>> _config = Config.from_yaml(config_path)

        >>> _config.email.sender = "foo@gmail.com"
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().host == EmailConnectionFactory.gmail_host
        True

        >>> _config.email.sender = "foo@outlook.com"
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().host == EmailConnectionFactory.outlook_host
        True

        >>> _config.email.sender = "foo@foobar.com"
        >>> _config.email.host = "smtp.foobar.com"
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().host == "smtp.foobar.com"
        True

        >>> _config.email.enable = False
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection() is None
        True
        """

        config = self.config.email

        if not config.enable:
            logger.info("Email send was disabled by config")
            return None

        sender = config.sender
        assert sender is not None
        domain = sender.split("@")[1]
        logger.debug("Sender domain: %s", domain)

        assert config.sender_password is not None
        password = config.sender_password.get_secret_value()

        # Handle Gmail emails
        if domain == "gmail.com":
            logger.debug("Gmail domain found")
            return EmailConnection(
                sender=sender,
                password=password,
                host=self.gmail_host,
                port=self.gmail_port,
                tls=self.gmail_tls,
            )

        # Handle Outlook emails
        elif domain.split(".")[0] in ("live", "outlook", "hotmail"):
            logger.debug("Outlook domain found")
            return EmailConnection(
                sender=sender,
                password=password,
                host=self.outlook_host,
                port=self.outlook_port,
                tls=self.outlook_tls,
            )

        # Handle custom domains
        else:
            logger.debug("Custom domain found")
            assert config.host is not None
            assert config.port is not None
            assert config.tls is not None

            return EmailConnection(
                sender=sender,
                password=password,
                host=config.host,
                port=config.port,
                tls=config.tls,
            )

    def get_connection_strict(self) -> EmailConnection:
        """
        Get an Email Connection, or raise if it would be null.
        """

        connection = self.get_connection()
        if connection is None:
            raise ValueError(
                "Connection is null! Probably you must set `email/enable: true` in your config file."
            )
        return connection


@define
class EmailExecutor:
    config: Config

    def are_messages_enabled(self) -> bool:
        """
        Return whether the email messages are enabled, so that:
        - `email/enable` is `true` in config file;
        - `email/recipient` is a valid non-null string;
        - an email connection can be established with sender's SMTP server
        """

        if not self.config.email.enable:
            return False

        if not self.config.email.recipient:
            return False

        factory = EmailConnectionFactory(config=self.config)

        connection = factory.get_connection()
        if not connection:
            return False

        try:
            connection.test_connection()
        except smtplib.SMTPException:
            return False
        else:
            return True

    def send_message(
        self, statuses: List[Status], *, subject: Optional[str] = None
    ) -> None:
        """
        Send a message with the content of object's `statuses`.

        It is also possible to change the subject of the message
        providing a non-null `subject`.

        Assumes that messages are enabled for the current session.
        """

        recipient: Optional[str] = self.config.email.recipient
        assert recipient is not None

        status_messages = [StatusMessage(status=status) for status in statuses]

        connection: EmailConnection = EmailConnectionFactory(
            config=self.config
        ).get_connection_strict()
        with connection:
            message = ExecutionMessage(
                sender=connection.sender,
                recipient=recipient,
                status_messages=status_messages,
            )
            if subject is not None:
                message.subject = subject
            connection.send_message(message=message)
            logger.info("Message sent correctly!")

    def send_mock_message(self, *, seed: int = 0) -> None:
        """
        Send a mock message in order to make the
        message display something relevant for the user.
        """

        random.seed(seed)

        fake = Faker(locale="en-US")
        fake.seed_instance(seed)

        statuses: List[Status] = []
        for profile in self.config.automsr.profiles:
            steps: List[Step] = []

            step_types: List[StepType] = StepType.get_ordered_steps()

            for step_type in step_types:
                if profile.skip:
                    outcome = OutcomeType.SKIPPED
                else:
                    outcome = random.choice(list(OutcomeType))

                if outcome is OutcomeType.SUCCESS:
                    explanation = None
                else:
                    explanation = fake.sentence(nb_words=4)

                if outcome is OutcomeType.SKIPPED:
                    duration = None
                else:
                    seconds = fake.random_number(digits=3, fix_len=True)
                    duration = datetime.timedelta(seconds=seconds)

                step = Step(
                    type=step_type,
                    outcome=outcome,
                    explanation=explanation,
                    duration=duration,
                )
                steps.append(step)

            points_should_be_null: bool = fake.pybool()
            if points_should_be_null:
                points = None
            else:
                points = fake.random_number(digits=6)

            status = Status(profile=profile, steps=steps, points=points)
            statuses.append(status)

        logger.info("Sending mock message...")
        self.send_message(statuses=statuses, subject="AutoMSR Mock Report message")
